
%{
open Ast

(* parser for tuple-types and expressions require at least two components *)
%}

%start program

%type <Ast.exp> exp exp2 exp3
%type <Ast.typ> typ typ2
%type <string*Ast.typ> typeAlias
%type <(string*Ast.typ) list> typeAliases
%type <((string*Ast.typ) list) * Ast.exp> program
%type <Ast.typ list> tupletyp
%type <Ast.exp list> explist


%token LLParen LRParen LColon LDot
%token LInt LArrow LRef
%token LTimes LPlus LGt
%token LFn LIf LThen LElse
%token LSet LGet LRef
%token LLet LIn LEq 
%token LTypeAlias LComma
%token LEof
%token <int> LConst
%token <string> LVar

%left LGt
%left LPlus
%left LTimes

%right LArrow
%%

typ:
  typ2 { $1 }
| tupletyp { TupleT($1) }

typ2:
  LInt { Int }
| typ2 LArrow typ2 { Arrow($1,$3) }
| LRef LLParen typ LRParen { Ref($3) }
| LLParen typ LRParen { $2 }
| LVar { Name($1) }

tupletyp:
  typ2 LTimes typ2      { [($1);($3)] }
| typ2 LTimes tupletyp { ($1)::$3 }

exp:
  exp2 { $1 }
| exp  LTimes exp { Times($1,$3) }
| exp  LPlus exp { Plus($1,$3) }
| exp  LGt   exp { GreaterThan($1,$3) }

exp2:
  exp3 { $1 }
| exp2 exp3 { Apply($1,$2) }

exp3:
  LLParen exp LRParen { $2 }
| LConst { Const($1) }
| LVar   { Var($1) } 
| LLParen LIf exp LThen exp LElse exp LRParen { If($3,$5,$7) }
| LLParen LFn  LVar LColon typ LDot exp LRParen { Fun($3,$5,$7) }
| LLParen exp  LSet exp LRParen { Set($2,$4) }
| LLParen LGet exp LRParen { Get($3) }
| LLParen LRef exp LRParen { Alloc($3) }
| LLParen LLet LVar LColon typ LEq exp LIn exp LRParen 
    { Apply(Fun($3,$5,$9),$7) }
| LLParen explist LRParen { Tuple($2) }
| exp3 LDot LConst { Field($1,$3) }
| LLParen exp LColon typ LRParen { Upcast($4,$2) }

explist:
  exp LComma exp { [($1);($3)] }
| exp LComma explist { ($1)::($3) }

typeAlias:
    LTypeAlias LVar LEq typ { ($2,$4) }

typeAliases:
    { [] }
| typeAlias typeAliases { ($1)::($2) }

program:
    typeAliases exp LEof { ($1,$2) }
